#!/usr/bin/python
# -*- coding: utf-8 -*-

from argparse import ArgumentParser
import numpy as np
import pandas as pd
import xarray as xr


class ReadOFormat(xr.Dataset):
    """reads Matrix from a file
    """
    _target_cols_zones = ['zone_no', 'zone_name']
    _target_cols_matrix = ['origins', 'destinations', 'values']

    def __init__(self,
                 zonefile,
                 matrixfile,
                 cols_zones=None,
                 cols_matrix=None,
                ):
        super(ReadOFormat, self).__init__()
        self.read_zones_csv(zonefile, cols_zones)
        self.read_matrix_csv(matrixfile, cols_matrix)

    def read_matrix_csv(self, filename, cols_matrix):
        """
        """
        target_cols = self._target_cols_matrix
        data_cols = cols_matrix
        pkey = target_cols[:2]

        da = self.read_file_to_da(data_cols, filename, target_cols, pkey)
        self['matrix'] = da['values']
        m = self['matrix']
        m.data[np.isnan(m.data)] = 0

    def read_zones_csv(self, filename, cols_zones):
        """
        """
        target_cols = self._target_cols_zones
        data_cols = cols_zones
        pkey = target_cols[0]
        col_name = target_cols[1]

        da = self.read_file_to_da(data_cols, filename, target_cols, pkey)
        dims = self._target_cols_zones[:1] + self._target_cols_matrix[:2]
        # set as origins and destinations
        for dim in dims:
            renamed_da = da.rename({pkey: dim})
            self[dim] = renamed_da[dim]
            name_dim = 'name_{}'.format(dim)
            self.coords[name_dim] = xr.IndexVariable(dims=[dim],
                                                     data=renamed_da[col_name])

    def read_file_to_da(self, data_cols, filename, target_cols, pkey):
        df = pd.read_csv(filename, usecols=data_cols)
        if not data_cols:
            # take the first columns of the table
            data_cols = df.columns.values[:len(target_cols)]

        # rename the columns to the target names
        msg = 'wrong number of columns specified'
        assert len(data_cols) == len(target_cols), msg
        rename_cols = dict(zip(data_cols,
                                target_cols))
        df.rename(columns=rename_cols, inplace=True)

        da = df.set_index(pkey).sort_index().to_xarray()
        return da



if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('-z', '--zonefile', dest='zonefile', required=True)
    parser.add_argument('-m', '--matrixfile', dest='matrixfile', required=True)
    parser.add_argument('--cols_zones', dest='cols_zones', nargs='*')
    parser.add_argument('--cols_matrix', dest='cols_matrix', nargs='*')

    options = parser.parse_args()
    ds = ReadOFormat(options.zonefile,
                     options.matrixfile,
                     options.cols_zones,
                     options.cols_matrix)
    print(ds)



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 21:00:21 2016

@author: MaxBohnet
"""

import pytest
import os
import numpy as np
from matrixconverters.read_ptv import ReadPTVMatrix
from matrixconverters.save_ptv import SavePTV


@pytest.fixture(scope='class')
def folder():
    return os.path.dirname(__file__)


@pytest.fixture(scope='class')
def matrix_fn(folder):
    fn = os.path.join(folder, 'matrix_v_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_out(folder):
    fn = os.path.join(folder, 'matrix.out')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_bk_out(folder):
    fn = os.path.join(folder, 'matrix_bk_format.out')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_bk(folder):
    fn = os.path.join(folder, 'matrix_bk_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_o(folder):
    fn = os.path.join(folder, 'matrix_o_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_e(folder):
    fn = os.path.join(folder, 'matrix_e_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_s(folder):
    fn = os.path.join(folder, 'matrix_s_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_or(folder):
    fn = os.path.join(folder, 'matrix_or_format.mtx')
    return fn


class TestReadPTV:
    """Test reading PTV matrices"""
    def test_01_read_v_format(self, matrix_fn):
        """Test reading v-Format"""
        ds = ReadPTVMatrix(filename=matrix_fn)
        self.print_matrix(ds)

    def test_02_read_bk_format(self, matrix_fn_bk):
        """Test reading bk-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bk)
        self.print_matrix(ds)

    def test_02a_read_o_format(self, matrix_fn_o, matrix_fn_or):
        """Test reading bk-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_o)
        self.print_matrix(ds)

        ds = ReadPTVMatrix(filename=matrix_fn_or)
        self.print_matrix(ds)

    def test_02b_read_e_s_format(self, matrix_fn_e, matrix_fn_s):
        """Test reading bk-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_e)
        self.print_matrix(ds)

        ds = ReadPTVMatrix(filename=matrix_fn_s)
        self.print_matrix(ds)

    def print_matrix(self, ds):
        print(ds)
        print('Histogram of travel times')
        print(np.histogram(ds.matrix.data, bins=range(0, 60, 10)))
        sum_before = ds.matrix.data.sum()
        print(sum_before)

    def test_03_save_v_format(self, matrix_fn_bk, matrix_fn_out):
        """Test writing v-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bk)
        print(np.histogram(ds.matrix.data, bins=range(0, 60, 10)))
        sum_before = ds.matrix.data.sum()
        print(sum_before)
        s = SavePTV(ds)
        s.savePTVMatrix(file_name=matrix_fn_out,
                        Ftype='VN', )
        ds2 = ReadPTVMatrix(filename=matrix_fn_out)
        print(np.histogram(ds2.matrix.data, bins=range(0, 60, 10)))
        sum_after = ds2.matrix.data.sum()
        print(sum_after)
        np.testing.assert_almost_equal(sum_before, sum_after, decimal=5)

    def test_04_save_bk_format(self, matrix_fn_bk, matrix_fn_bk_out):
        """Test writing bk-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bk)
        print(np.histogram(ds.matrix.data, bins=range(8)))
        sum_before = ds.matrix.data.sum()
        print(sum_before)
        s = SavePTV(ds)
        s.savePTVMatrix(file_name=matrix_fn_bk_out,
                        Ftype='BK', )
        ds2 = ReadPTVMatrix(filename=matrix_fn_bk_out)
        print(np.histogram(ds2.matrix.data, bins=range(8)))
        sum_after = ds2.matrix.data.sum()
        print(sum_after)
        assert sum_before == sum_after



# -*- coding: utf-8 -*-
# Name:        saveArray
# Purpose:
#
# Author:      Max Bohnet
#
# Created:     21/09/2011
# Copyright:   (c) Max Bohnet 2011


import gzip
import numpy as np


def writeFormattedLine(f, text, length):
    f.write(text.ljust(length) + b"\r\n")


class SavePTV(object):

    def __init__(self, ds):
        """"""
        self.ds = ds

    def savePTVMatrix(self, file_name, Ftype, version=0.11,
                      zipped=False):
        """
        save array im PTV-Format

        Parameters
        ----------
        file_name : string
            output file_name
        FType : {'B', 'V'}
            Type of PTV-File Format
            either als binary file ("B")
            or as V-Format ("V").
        version : srting
            version number
        zipped : bool
            To save Disc Space and increase speed,
            the file can be saved in a gzip-packed format (zipped = True)
        """
        if zipped:
            OpenFile = gzip.open
            fn = file_name + ".gzip"
        else:
            OpenFile = open
            fn = file_name

        if Ftype.startswith("B"):
            self.write_format_b(OpenFile, fn, version)
        elif Ftype.startswith("V"):
            self.write_format_v(OpenFile, fn, Ftype)

    def write_format_v(self, OpenFile, fn, Ftype):
        dtypes = {'float32': "Y4", 'float64': "Y5",
                  'int32': "Y3", 'int16': "Y2"}
        outformat = {'float32': "%0.3f", 'float64': "%0.6f",
                     'int32': "%d", 'int16': "%d"}
        with OpenFile(fn, "w") as f:
            anz_bez_listen = getattr(self.ds.attrs, 'AnzBezeichnerlisten', 1)
            ZeitVon = getattr(self.ds.attrs, 'ZeitVon', 0)
            ZeitBis = getattr(self.ds.attrs, 'ZeitBis', 24)
            Faktor = getattr(self.ds.attrs, 'Faktor', 1)
            VMAktKennung = getattr(self.ds.attrs, 'VMAktKennung', 0)
            dtype = self.ds.matrix.dtype
            f.write("$%s, %s\n" % (Ftype, dtypes[str(dtype)]))
            if "M" in Ftype:
                f.write("* Verkehrsmittelkennung:\n %d \n" % VMAktKennung)
            if "N" not in Ftype:
                f.write("* Zeitintervall:\n {v} {b} \n".format(
                    v=ZeitVon,
                    b=ZeitBis))
                f.write("* Faktor:\n %d \n" % Faktor)
            zone_no = self.ds.zone_no.data
            n_zones = len(zone_no)
            f.write("* Anzahl Bezirke:\n %d\n" % n_zones)
            f.write("* Bezirksnummern\n")
            for i in range(0, n_zones, 10):
                f.write("\t".join(map(lambda x:
                                      str(x),
                                      zone_no[i: i + 10])) + "\n")
            f.write("* Matrixwerte\n")
            values = self.ds.matrix.data
            fmt = outformat[str(dtype)]
            for i in range(n_zones):
                no = zone_no[i]
                f.write("* %s \n" % no)
                for j in range(0, n_zones, 10):
                    values[i, j: j + 10].tofile(f, sep=' ', format=fmt)
                    f.write('\n')

            if 'zone_names' in self.ds:
                zone_names = self.ds.zone_names.data
                fmt = '{no} "{name}"\n'
                for i in range(n_zones):
                    f.write(fmt.format(no=zone_no[i], name=zone_names[i]))

    def write_format_b(self, OpenFile, fn, version):
        m = self.ds.matrix.data
        with OpenFile(fn, "wb") as f:
            anz_bez_listen = getattr(self.ds.attrs, 'AnzBezeichnerlisten', 1)
            ZeitVon = getattr(self.ds.attrs, 'ZeitVon', 0)
            ZeitBis = getattr(self.ds.attrs, 'ZeitBis', 24)
            Faktor = getattr(self.ds.attrs, 'Faktor', 1)
            VMAktKennung = getattr(self.ds.attrs, 'VMAktKennung', 0)
            writeFormattedLine(f, b"PTVSYSTEM   MUULI       VMatrixComp1", 44)
            writeFormattedLine(f, b"VMatversionsnummer: %0.2f" % version, 28)
            writeFormattedLine(f, b"Anzahl Dimensionen: %d" % m.ndim, 28)
            writeFormattedLine(f, b"Seitenlaengen der Matrix:", 30)
            text = b""
            for n in range(m.ndim):
                text += (b"%d" % m.shape[n]).ljust(8)
            writeFormattedLine(f, text, 48)
            writeFormattedLine(f, b"% -30s%d" % (b"Anzahl Bezeichnerlisten:",
                                                 anz_bez_listen),
                               38)
            writeFormattedLine(f, b"% -20s%02d:00   %02d:00" % (
                b"Zeitbereich:",
                ZeitVon,
                ZeitBis,
                ),
                               36)
            writeFormattedLine(f, b"% -20s%s" % (b"RandomRound:", b"Nein"), 28)
            writeFormattedLine(f, b"% -24s%0.2f" % (b"Faktor:",
                                                    Faktor), 28)
            for i in range(20):
                writeFormattedLine(f, b" ", 80)
            f.write(np.array(6682, dtype="i2").tostring())  # header
            f.write(b" " * 80)
            f.write(np.array(11, dtype="i2").tostring())  # header
            f.write(np.array(m.ndim, dtype="i2").tostring())  # Dimensions
            # Zeilen je Dimension
            f.write(np.array(m.shape[-2], dtype="i4").tostring())
            f.write(np.array(2080, dtype="i4").tostring())  # header
            # Spalten
            f.write(np.array(m.shape[-1], dtype="i4").tostring())
            f.write(np.array(2080, dtype="i4").tostring())  # ???
            if m.ndim == 3:
                # Bloecke
                f.write(np.array(m.shape[0], dtype="i4").tostring())
                f.write(np.array(2080, dtype="i4").tostring())  # ???

            if m.dtype.char == 'f':
                # 3 Datentyp float
                f.write(np.array(3, dtype="i2").tostring())
                typecode = '<f4'
            elif m.dtype.char == 'd':
                # 4 Datentyp double
                f.write(np.array(4, dtype="i2").tostring())
                typecode = '<f8'
            else:
                # z.B. integer oder unsigned integer:
                # 3 Datentyp float
                # ToDo: Prüfen!!!
                f.write(np.array(3, dtype="i2").tostring())
                typecode = '<f4'
            # AnzBezeichnerlisten, was immer das auch ist ???
            f.write(np.array(anz_bez_listen, dtype="i2").tostring())
            f.write(np.array(ZeitVon, dtype="f4").tostring())  # ZeitVon
            f.write(np.array(ZeitBis, dtype="f4").tostring())  # ZeitBis
            # VMAktKennung
            f.write(np.array(VMAktKennung, dtype="i4").tostring())
            # Unbekannt U als float
            f.write(np.array(Faktor, dtype="<f4").tostring())
            # zones
            f.write(np.array(self.ds.zone_no.data).astype("i4").tostring())
            f.write(m.astype(typecode).flatten().tostring())  # Matrix

    def savePSVMatrix(self, fileName, ftype="CC", maxWidth=1000):
        """ exports array in PSV-Format
        """
        m = self.ds.matrix.data
        if not m.ndim == 2:
            raise ValueError('matrix has to have 2 dimensions')
        rows, cols = m.shape
        if not hasattr(self, 'zones'):
            self.zones = range(1, rows + 1)
        f = open(fileName, "w")
        try:
            Za = max(rows, cols)
            Zi = min(rows, cols)
            f.write("%s; Za %s; Zi %s;\n" % (ftype, Za, Zi))
            if ftype == "CC":
                for i in range(rows):
                    # teste, ob Zeile nicht nur aus Nullen besteht
                    if m[i].any():
                        # addiere 1, um Zellennummer 0 zu vermeiden
                        row = str(m.zones[i])
                        for j in range(cols):
                            if len("%s %s %s" % (row, m.zones[j],
                                                 m[i, j])) >= maxWidth:
                                row += "\n"
                                f.write(row)
                                row = str(m.zones[i])
                            if m[i, j] != 0:
                                row += " %s %s " % (m.zones[j], m[i, j])
                        row += "\n"
                        f.write(row)

            elif ftype == "CN":
                for i in range(rows):
                    # teste, ob Zeile nicht nur aus Nullen besteht
                    if m[i].any():
                        row = str(m.zones[i])
                        for j in range(cols):
                            if len("%s %s" % (row, m[i, j])) >= maxWidth:
                                row += "\n"
                                f.write(row)
                                row = str(m.zones[i])
                            row += " " + str(m[i, j])
                        row += "\n"
                        f.write(row)
            else:
                raise TypeError("Kann Typ %s nicht schreiben" % ftype)
        finally:
            f.close()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import gzip
import zlib
from argparse import ArgumentParser
from collections import deque
import numpy as np
import xarray as xr


class ReadPTVMatrix(xr.Dataset):
    """reads PTV-Matrix from file
    the matrix-type is determined by the header
    reades $M, $V, $O, $E, and the binary types $BI and $B

    Arguments:
        FileName: path to file on disk
        header: length ot the header
        zipped: true, if file is zipped
    """
    def __init__(self, filename, header=2048, zipped=False):
        super(ReadPTVMatrix, self).__init__()
        self.attrs['fn'] = filename
        self.attrs['ZeitVon'] = 0
        self.attrs['ZeitBis'] = 0
        self.attrs['Faktor'] = 1
        self.attrs['VMAktKennung'] = 0
        self.attrs['AnzBezeichnerlisten'] = 1

        if filename.endswith(".gzip") or zipped:
            self.attrs['open'] = gzip.open
        else:
            self.attrs['open'] = open
        with self.openfile(mode='rb') as f:
            line = f.readline().strip()

        if line.startswith(b"$M"):
            self.readPTVMatrixM()
        elif line.startswith(b"$V"):
            self.readPTVMatrixV()
        elif line.startswith(b"$O"):
            self.readPTVMatrixO()
        elif line.startswith(b"$E"):
            self.readPTVMatrixE()
        elif line.startswith(b"$S"):
            self.readPTVMatrixE()
        elif line.strip(bytes([3, 0])).startswith(b"$BI"):
            self.readPTVMatrixBI()
        elif line.strip(bytes([3, 0])).startswith(b"$BK"):
            self.readPTVMatrixBK()
        elif line.strip(bytes([3, 0])).startswith(b"$B"):
            self.readPTVMatrixB()

        else:
            self.readPTVMatrixB(header)

        del self.attrs['open']

    def readPTVMatrixO(self):
        with self.openfile(mode="r") as f:
            rows = self.read_header(f)
        loc = self.matrix.loc
        for row in rows:
            r = row.split()
            fr = int(r[0])
            to = int(r[1])
            value = float(r[2])
            loc[fr, to] = value

    def readPTVMatrixE(self):
        with self.openfile(mode="r") as f:
            rows = self.read_header(f)
        loc = self.matrix.loc
        for row in rows:
            cols = row.lstrip('-').split()
            fr = int(cols[0])
            for i in range(1, len(cols), 2):
                to = int(cols[i])
                value = float(cols[i + 1])
                loc[fr, to] = value

    def read_header(self, f):
        line = f.readline()
        MatrixTyp = line.split("$")[-1].split(";")[0]
        if "M" in MatrixTyp:
            self.attrs['VMAktKennung'] = self.readWert(f)

        if "N" not in MatrixTyp:
            ZeitVon, ZeitBis, Faktor = self.readWerte(f, 3)
            self.attrs['ZeitVon'] = ZeitVon
            self.attrs['ZeitBis'] = ZeitBis
            self.attrs['Faktor'] = Faktor

        rows, line = self.read_values_in_o_format_to_list(f)
        self.read_names_o_format(f, line)
        return rows

    def openfile(self, mode='r'):
        return self.attrs['open'](self.attrs['fn'], mode=mode)

    def readPTVMatrixV(self):
        with self.openfile() as f:
            line = f.readline()
            MatrixTyp = line.split("$")[-1].split(";")[0]
            if not MatrixTyp.startswith("V"):
                print("Keine Matrix im V-Format!")
                raise TypeError
            if "M" in MatrixTyp:
                self.attrs['VMAktKennung'] = self.readWert(f)

            if "N" not in MatrixTyp:
                ZeitVon, ZeitBis, Faktor = self.readWerte(f, 3)
                self.attrs['ZeitVon'] = ZeitVon
                self.attrs['ZeitBis'] = ZeitBis
                self.attrs['Faktor'] = Faktor

            n_zones = self.readWert(f)
            self.create_zones(n_zones)
            self.read_values_to_array(f, self.zone_no)

            self.create_matrix(n_zones)
            self.read_values_to_array(f, self.matrix)
            self.create_zone_names(n_zones)
            self.read_names(f, self.zone_name)

    def create_zone_names(self, n_zones, name='zone_name'):
        self[name] = xr.DataArray(
            np.empty((n_zones, ), dtype='O'),
            coords=(self.zone_no.data, ),
            dims=('zone_no',),
            name=name,)

    def create_matrix(self, n_zones, dtype='f8'):
        self['matrix'] = xr.DataArray(
            np.empty((n_zones, n_zones), dtype=dtype),
            coords=(self.zone_no, self.zone_no),
            dims=('origins', 'destinations'),
            name='matrix',)

    def create_zones(self, n_zones, name='zone_no'):
        self.coords[name] = xr.DataArray(np.empty((n_zones, ), dtype='i4'),
                                         dims=('zones',),)

    def read_names(self, f, arr):
        """Read the zone names"""
        line = f.readline()
        while line.startswith("*") or line == "\n":
            line = f.readline()
        if line:
            if line.upper().startswith('$NAMES'):

                line = f.readline().strip()
                while line:
                    l = line.split(' "')
                    zone_no = int(l[0])
                    name = l[1].strip('"')
                    arr.loc[zone_no] = name
                    line = f.readline().strip()

    def read_names_o_format(self, f, line):
        """Read the zone names"""
        names = deque()
        while line.startswith("*") or line == "\n":
            line = f.readline()
        if line:
            if line.upper().startswith('$NAMES'):

                line = f.readline().strip()
                while line:
                    l = line.split(' "')
                    zone_no = int(l[0])
                    name = l[1].strip('"')
                    names.append((zone_no, name))
                    line = f.readline().strip()
        n_zones = len(names)
        self.create_zones(n_zones)
        for i, (zone_no, name) in enumerate(names):
            self.zone_no.data[i] = zone_no
        self.create_zone_names(n_zones)
        for i, (zone_no, name) in enumerate(names):
            self.zone_name.loc[zone_no] = name
        self.create_matrix(n_zones)

    def readPTVMatrixBK(self):
        """
        Die Länge des Headers stehen in Byte 5/6
        Zuerst wird mit Hilfe dieses offsets die Anzahl der Zellen ausgelesen
        und eine leere Matrix gebildet
        dann werden Verkehrsmittelkennung gelesen.
        Dann folgen die Zellennummern
        Schlieslich wird für jede Zeile der Matrix zuerst
        die Länge der gepakten Daten dieser Zeile (lenChunk) gelesen
        und dann die Daten mit zlib.decompress entpackt
        und mit np.fromstring in ein Zeilen-array verwandelt,
        Damit wird die Matrix zeilenweise "gefüllt"
        Zwischen den Zeilen befinden sich jeweils 16 byte,
        deren Sinn unklar ist (ggf. eine Prüfsumme?)
        Diese werden ignoriert.
        """
        with self.openfile(mode="rb") as f:
            f.seek(5, 0)
            header = np.fromstring(f.read(2), dtype="i2")[0]
            f.seek(header, 0)
            f.seek(23 - 16, 1)
            self.attrs['VMAktKennung'] = np.fromstring(f.read(2),
                                                       dtype="i2")[0]
            f.seek(14, 1)
            n_zones = np.fromstring(f.read(4), dtype="i4")[0]
            self.create_zones(n_zones)

            data_type = np.fromstring(f.read(2), dtype="i2")[0]
            f.seek(1, 1)
            if data_type == 4:
                dtype = '<f4'
            elif data_type == 5:
                dtype = '<f8'
            self.create_matrix(n_zones, dtype=dtype)

            n_zones2 = np.fromstring(f.read(4), dtype="i4")[0]
            self.zone_no.data[:] = np.fromstring(
                f.read(n_zones * 4), dtype="i4")
            self.create_zones(n_zones2, name='zone_no2')
            self.zone_no2.data[:] = np.fromstring(
                f.read(n_zones * 4), dtype="i4")
            self.create_zone_names(n_zones)
            for i in range(n_zones):
                Zeichen = np.fromstring(f.read(4), dtype="i4")[0]
                self.zone_name.data[i] = f.read(Zeichen * 2).decode('utf16')

            self.create_zone_names(n_zones, name='zone_names2')
            for i in range(n_zones2):
                Zeichen = np.fromstring(f.read(4), dtype="i4")[0]
                self.zone_names2.data[i] = f.read(Zeichen * 2).decode('utf16')
            f.seek(1, 1)
            self.attrs['Unbek1'] = np.fromstring(f.read(8), dtype="f8")[0]
            for i in range(n_zones):
                lenChunk = np.fromstring(f.read(4), dtype="i4")[0]
                unpacked = zlib.decompress(f.read(lenChunk))
                self.matrix[i] = np.fromstring(unpacked, dtype=dtype)
                f.seek(16, 1)

    def readPTVMatrixBI(self):
        """
        Die Länge des Headers stehen in Byte 5/6
        Zuerst wird mit Hilfe dieses offsets die Anzahl der Zellen ausgelesen
        und eine leere Matrix gebildet
        dann werden Verkehrsmittelkennung gelesen.
        Dann folgen die Zellennummern
        Schlieslich wird für jede Zeile der Matrix
        zuerst die Länge der gepakten Daten dieser Zeile (lenChunk) gelesen
        und dann die Daten mit zlib.decompress entpackt
        und mit np.fromstring in ein Zeilen-array verwandelt,
        Damit wird die Matrix zeilenweise "gefüllt"
        Zwischen den Zeilen befinden sich jeweils 16 byte,
        deren Sinn unklar ist (ggf. eine Prüfsumme?)
        Diese werden ignoriert.
        """
        with self.openfile("rb") as f:
            f.seek(5, 0)
            header = np.fromstring(f.read(2), dtype="i2")[0]
            f.seek(header, 0)
            f.seek(23 - 16, 1)
            self.attrs['VMAktKennung'] = np.fromstring(
                f.read(2), dtype="i2")[0]
            f.seek(14, 1)
            n_zones = np.fromstring(f.read(4), dtype="i4")[0]
            data_type = np.fromstring(f.read(2), dtype="i2")[0]
            if data_type == 4:
                dtype = '<f4'
            elif data_type == 5:
                dtype = '<f8'
            self.create_zones(n_zones)
            self.create_matrix(n_zones, dtype=dtype)
            f.seek(1, 1)
            self.zone_no.data[:] = np.fromstring(
                f.read(4 * n_zones), dtype="i4")[0]
            f.seek(9, 1)
            m = self.matrix.data

            for i in range(n_zones):
                lenChunk = np.fromstring(f.read(4), dtype="i4")[0]
                unpacked = zlib.decompress(f.read(lenChunk))
                m[i] = np.fromstring(unpacked, dtype=dtype)
                m.seek(16, 1)

    def readPTVMatrixB(self, header=2048):
        with self.openfile(mode="rb") as f:
            f.seek(header)
            f.seek(2, 1)
            Dimensions = np.fromstring(f.read(2), dtype="i2")[0]
            MatrixZeilen = np.fromstring(f.read(4), dtype="i4")[0]
            f.seek(4, 1)
            MatrixSpalten = np.fromstring(f.read(4), dtype="i4")[0]
            f.seek(4, 1)
            if Dimensions == 3:
                MatrixBlock = np.fromstring(f[header + 20:header + 24],
                                            dtype="i4")[0]
                f.seek(4, 1)
                shape = (MatrixBlock, MatrixZeilen, MatrixSpalten)
            elif Dimensions == 2:
                MatrixBlock = 1
                shape = (MatrixZeilen, MatrixSpalten)
            else:
                msg = 'only 2d or 3d dimensions allowed, found %s dimensions'
                raise ValueError(msg % Dimensions)
            data_type = np.fromstring(f.read(2), dtype="i2")
            if data_type == 3:
                data_length = 4
                dtype = '<f4'
            elif data_type == 4:
                data_length = 8
                dtype = '<f8'
            AnzBezeichnerlisten = np.fromstring(f.read(2),
                                                dtype="i2")
            AnzFelder = MatrixZeilen * MatrixSpalten * MatrixBlock

            n_zones = max(MatrixZeilen, MatrixSpalten)
            self.create_zones(n_zones)
            self.attrs['ZeitVon'] = int(np.fromstring(f.read(4), dtype="f4"))
            self.attrs['ZeitBis'] = int(np.fromstring(f.read(4), dtype="f4"))
            self.attrs['VMAktKennung'] = int(
                np.fromstring(f.read(4), dtype="i4"))
            self.attrs['Faktor'] = int(np.fromstring(f.read(4), dtype="f4"))
            if AnzBezeichnerlisten > 0:
                self.zone_no.data[:] = np.fromstring(
                    f.read(4 * n_zones), dtype='i4')
            self.create_matrix(n_zones, dtype=dtype)
            arr = np.fromstring(f.read(AnzFelder * data_length), dtype=dtype)
            self.matrix.data[:] = arr.reshape(shape)

    def readWert(self, f):
        line = f.readline()
        while line.startswith("*") or line == "\n":
            line = f.readline()
        Wert = int(line.strip())
        return Wert

    def read_values_to_array(self, f, arr, sep=' '):
        """read values to a numpy array at pos x"""
        flat_arr = arr.data.ravel()
        n_total = len(flat_arr)
        pos_from = 0
        while pos_from < n_total:
            line = f.readline()
            while line.startswith("*"):
                line = f.readline()
            row = np.fromstring(line, sep=sep, dtype=arr.dtype)
            pos_to = pos_from + len(row)
            flat_arr[pos_from:pos_to] = row
            pos_from = pos_to

    def read_values_in_o_format_to_list(self, f):
        """read values to a numpy array at pos x"""
        rows = deque()
        for line in f:
            if line.startswith("*") or line == '\n':
                continue
            if line.startswith('$'):
                return rows, line
            row = line.strip()
            rows.append(row)
        return rows, line

    def readWerte(self, f, AnzWerte):
        Werte = []
        WerteGefunden = 0
        while WerteGefunden < AnzWerte:
            line = f.readline()
            while line.startswith("*"):
                line = f.readline()
            Werte += map(lambda x: float(x), line.strip().split())
            WerteGefunden = len(Werte)
        return Werte

if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('-f', dest='fn')
    options = parser.parse_args()
    ds = ReadPTVMatrix(options.fn)
